/*

	mplayerplug-in

	This is a plugin for Netscape and Mozilla that uses mplayer to play supported videoes from websites
	It handles embedded and full page files equally.

	License: GPL

	Created 2003-01-09

	version 0.95	2003-10-01
	Version 0.91	2003-09-10
	Version 0.90	2003-09-07

	Authors:
	Kevin DeKorte	kdekorte@yahoo.com

 	Contributors
	Fabian Franz	gmplayer patches and support
	Todd Kirby	embed scale aspect option
	Ed Fisher	novop option
	Don Estberg	post processing options and QT Streams

	Code inspired by the webfreetv.com plugin http://www.webfreetv.com/linux/ by markus.rechberger@bluebull.com

 *
 */

#include "mplayerplug-in.h"

// This routine is only called when the plugin library is newer than the pluginrc file

char *NPP_GetMIMEDescription(void)
{

    char *ret;
    char MimeTypes[4000], line[1000], parse[1000], buffer[1000],
	codecname[32], customline[255];
    FILE *config, *customtypes;
    int i, use_custom_mime_types;


    enable_real = 0;
    enable_qt = 1;
    enable_wm = 1;
    enable_mpeg = 1;
    use_custom_mime_types = 0;

    DEBUG = 1;
    if (DEBUG)
	printf("Reading config file for codecs\n");
    // clear the buffer
    for (i = 0; i < 4000; i++) {
	MimeTypes[i] = '\0';
    }

    config = NULL;

    if (config == NULL) {
	sprintf(buffer, "%s", getenv("HOME"));
	strcat(buffer, "/.mplayer/mplayerplug-in.conf");
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	sprintf(buffer, "%s", getenv("HOME"));
	strcat(buffer, "/.mozilla/mplayerplug-in.conf");
	config = fopen(buffer, "r");
    }

    if (config == NULL) {
	config = fopen("/etc/mplayerplug-in.conf", "r");
    }

    if (config == NULL) {
	// no config file
    } else {
	while (fgets(buffer, 1000, config) != NULL) {

	    if (DEBUG)
		printf("Buffer: %s\n", buffer);

	    if (strncasecmp(buffer, "enable-qt", 9) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_qt);
		if (DEBUG)
		    printf("qt:%i\n", enable_qt);
		continue;
	    }
	    // Windows Media
	    if (strncasecmp(buffer, "enable-wm", 9) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_wm);
		if (DEBUG)
		    printf("wm:%i\n", enable_wm);
		continue;
	    }
	    // RealMedia
	    if (strncasecmp(buffer, "enable-real", 11) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_real);
		if (DEBUG)
		    printf("real:%i\n", enable_real);
		continue;
	    }
	    // MPEG
	    if (strncasecmp(buffer, "enable-mpeg", 11) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &enable_mpeg);
		if (DEBUG)
		    printf("mpeg:%i\n", enable_mpeg);
		continue;
	    }

	    if (strncasecmp(buffer, "use-mimetypes", 13) == 0) {
		sprintf(parse, "%s", strtok(buffer, "="));
		sprintf(parse, "%s", strtok(NULL, "="));
		sscanf(parse, "%i", &use_custom_mime_types);
		if (DEBUG)
		    printf("custom mimetypes:%i\n", use_custom_mime_types);
		continue;
	    }


	}
	fclose(config);
    }

    if (use_custom_mime_types == 1) {
	customtypes = NULL;

	if (customtypes == NULL) {
	    sprintf(buffer, "%s", getenv("HOME"));
	    strcat(buffer, "/.mplayer/mplayerplug-in.types");
	    customtypes = fopen(buffer, "r");
	}

	if (customtypes == NULL) {
	    sprintf(buffer, "%s", getenv("HOME"));
	    strcat(buffer, "/.mozilla/mplayerplug-in.types");
	    config = fopen(buffer, "r");
	}

	if (customtypes == NULL) {
	    customtypes = fopen("/etc/mplayerplug-in.types", "r");
	}

	if (customtypes != NULL) {
	    while (fgets(customline, 255, customtypes)) {
		strcat(MimeTypes, customline);
	    }
	    fclose(customtypes);
	}


    } else {

	if (enable_qt == 1) {
	    strcat(MimeTypes,
		   "video/quicktime:mov:Quicktime;"
		   "video/x-quicktime:mov:Quicktime;"
		   "video/quicktime:mp4:Quicktime;"
		   "video/quicktime:sdp:Quicktime - Session Description Protocol;"
		   "application/x-quicktimeplayer:mov:Quicktime;");
	}

	if (enable_wm == 1) {
	    strcat(MimeTypes,
		   "video/x-ms-asf-plugin:asf,asx:Windows Media;"
		   "video/x-msvideo:avi:AVI;"
		   "video/msvideo:avi:AVI;"
		   "application/x-mplayer2:wmv:WMV;"
		   "video/x-ms-wm:asf:MSNBCPlayer;"
		   "video/x-ms-asf:asf,asx:Windows Media;"
		   "video/x-ms-wmv:wmv:Windows Media;"
		   "video/x-ms-wvx:wvx:Windows Media;"
		   "audio/x-ms-wax:wax:Windows Media;"
		   "audio/x-ms-wma:wma:Windows Media;"
		   "application/x-drm-v2:asx:Windows Media;"
		   "audio/wav:wav:Microsoft wave file;"
		   "audio/x-wav:wav:Microsoft wave file;");
	}


	if (enable_real == 1) {
	    strcat(MimeTypes,
		   "audio/x-pn-realaudio:ram,rm:RealAudio;"
		   "audio/x-pn-realaudio-plugin:rpm:RealAudio;"
		   "audio/x-realaudio:ra:RealAudio;");
	}

	if (enable_mpeg == 1) {
	    strcat(MimeTypes,
		   "video/mpeg:mpg,mpeg:MPEG;"
		   "audio/mpeg:mpg,mpeg:MPEG;"
		   "video/x-mpeg:mpg,mpeg:MPEG;"
		   "video/x-mpeg2:mpv2,mp2ve:MPEG2;"
		   "audio/mpeg:mpg,mpeg:MPEG;"
		   "audio/x-mpeg:mpg,mpeg:MPEG;"
		   "audio/mpeg2:mp2:MPEG audio;"
		   "audio/x-mpeg2:mp2:MPEG audio;"
		   "audio/mpeg3:mp3:MPEG audio;"
		   "audio/x-mpeg3:mp3:MPEG audio;"
		   "audio/mp3:mp3:MPEG audio;"
		   "video/mp4:mp4:MPEG 4 Video;");
	}
	// FLI
	strcat(MimeTypes,
	       "video/fli:fli,flc:FLI animation;"
	       "video/x-fli:fli,flc:FLI animation;");

        // Vivo
	strcat(MimeTypes,
	       "video/vnd.vivo:viv,vivo:VivoActive;");

    }
    ret = (char *) malloc(strlen(MimeTypes));
    MimeTypes[strlen(MimeTypes)-1] = '\0';
    memcpy(ret, MimeTypes, strlen(MimeTypes));
    if (DEBUG)
	printf("%s\n", ret);
    DEBUG = 0;
    return ret;

}

/* This function returns the decription to Mozilla */

NPError NPP_GetValue(void *future, NPPVariable variable, void *value)
{
    NPError err = NPERR_NO_ERROR;

    switch (variable) {
    case NPPVpluginNameString:
	/* have to have these names in the name string to make Atom Films work */
	*((const char **) value) =
	    "mplayerplug-in v1.0pre2 handles QuickTime Windows Media Player Plugin";
	break;
    case NPPVpluginDescriptionString:
	*((const char **) value) =
	    "Video Player Plug-in for QuickTime and Windows Media Player streams using mplayer";
	break;
    default:
	err = NPERR_GENERIC_ERROR;
    }
    return err;
}

NPError NPP_Initialize(void)
{
    return NPERR_NO_ERROR;
}


jref NPP_GetJavaClass()
{
    return NULL;
}

void NPP_Shutdown(void)
{
}


NPError NPP_New(NPMIMEType pluginType,
		NPP instance,
		uint16 mode,
		int16 argc, char *argn[], char *argv[],
		NPSavedData * saved)
{
    PluginInstance *This;
    int i;
    char parse[1000];
    char logfile[1000];
    char *cp;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
    This = (PluginInstance *) instance->pdata;
    InitPrivateData(instance);

    DESTROYED = 0;

    This->mode = mode;
    LoadConfigFile(instance);

    if (mode == NP_EMBED) {

	if (DEBUG)
	    fprintf(This->logfp, "Embedded mode\n");
	for (i = 0; i < argc; i++) {

	    if (DEBUG) {
		fprintf(This->logfp, "Argument Name: %s\n", argn[i]);
		fprintf(This->logfp, "Argument Value: %s\n", argv[i]);
	    }

	    if (strncasecmp(argn[i], "src", 3) == 0) {
		This->url = strdup(argv[i]);
		This->state = STATE_HAVEURL;
	    }

            if (strncasecmp(argn[i], "filename", 8) == 0) {
                This->fname = strdup(argv[i]);
                This->state = STATE_HAVEURL;
                NPN_GetURL(instance,This->fname,NULL);
            }

	    if ((strncasecmp(argn[i], "href", 4) == 0)
		|| (strncasecmp(argn[i], "qtsrc", 5) == 0)) {
		This->href = strdup(argv[i]);
		This->state = STATE_HAVEURL;
	    }

	    if (strncasecmp(argn[i], "height", 6) == 0) {
		sscanf(argv[i], "%i", &This->embed_height);
	    }

	    if (strncasecmp(argn[i], "width", 5) == 0) {
		sscanf(argv[i], "%i", &This->embed_width);
	    }

	    /* handle 'scale' attribute used by QT Plugin */
	    if (strncasecmp(argn[i], "scale", 5) == 0) {
		if (strncasecmp(argv[i], "aspect", 6) == 0) {
		    This->maintain_aspect = 1;
		}
	    }

	    if (strncasecmp(argn[i], "loop", 4) == 0) {

		lowercase(argv[i]);
		if (DEBUG)
		    fprintf(This->logfp, "argv[i]=%s\n", argv[i]);

		if (strstr(argv[i], "true") || strstr(argv[i], "yes")) {
		    This->loop = 1;
		} else {
		    This->loop = 0;
		}
		if (DEBUG)
		    fprintf(This->logfp, "loop=%i\n", This->loop);
	    }

	    if (This->nQtNext < 256
		&& (strncasecmp(argn[i], "qtnext", 6) == 0)
		&& argv[i][0] == '<') {
		sprintf(parse, "%s", strtok(&argv[i][1], ">"));
		if ((cp = strchr(parse, ' '))
		    && strlen(parse) == cp - parse + 1)
		    *cp = (char) NULL;
		This->qtNext[This->nQtNext++] = strdup(parse);
		sprintf(parse, "%s", strtok(NULL, "<"));
		if (strcmp(parse, "T")) {
		    if (DEBUG)
			fprintf(This->logfp,
				"qtNext%i expected \"T\" found \"%s\"\n",
				This->nQtNext, parse);
		    This->nQtNext--;
		} else {
		    sprintf(parse, "%s", strtok(NULL, ">\n"));
		    if (strcmp(parse, "myself")) {
			if (DEBUG)
			    fprintf(This->logfp,
				    "qtNext%i expected \"myself\" found \"%s\"\n",
				    This->nQtNext, parse);
			This->nQtNext--;
		    } else if (DEBUG)
			fprintf(This->logfp, "qtNext%i=%s\n",
				This->nQtNext,
				This->qtNext[This->nQtNext - 1]);
		}
	    }



	}
    } else {
	if (DEBUG)
	    fprintf(This->logfp, "New, full mode %i\n", This->mode);
    }


    if (This != NULL)
	return NPERR_NO_ERROR;
    else
	return NPERR_OUT_OF_MEMORY_ERROR;
}


NPError NPP_Destroy(NPP instance, NPSavedData ** save)
{

    char cmd[1024];
    PluginInstance *This;
    int ret;

    if (DEBUG)
	printf("destroy %p\n", instance);

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    DESTROYED = 1;
    This = (PluginInstance *) instance->pdata;
    if (This != NULL) {
	if (DEBUG)
	    fprintf(This->logfp, "\nIn NPP_Destroy\n");

	if (DEBUG)
	    ret = fclose(This->logfp);
    }
    /* cleanup streams and files */
    pthread_attr_destroy(&(This->thread_attr));
    pthread_mutex_destroy(&(This->playlist_mutex));
    pthread_cond_destroy(&(This->playlist_complete_cond));

    if (This != NULL) {
	if (DEBUG)
	    printf("Freeing memory\n");

	if (This->td != NULL) {
	    NPN_MemFree(This->td);
	    This->td = NULL;
	}
	NPN_MemFree(instance->pdata);
	This = NULL;
	instance->pdata = NULL;
    }

    return NPERR_NO_ERROR;

}



NPError NPP_SetWindow(NPP instance, NPWindow * window)
{
    PluginInstance *This;
    Widget netscape_widget, screen;
    NPSetWindowCallbackStruct *ws;
    int ret, i;
    char *buffer, *buffer2;
    Node *n;

    if (instance == NULL) {
	if (DEBUG)
	    printf("SetWindow Called with invalid instance\n");
	return NPERR_INVALID_INSTANCE_ERROR;
    }

    This = (PluginInstance *) instance->pdata;

    if ((window == NULL) || (window->window == NULL)) {
	return NPERR_NO_ERROR;
    }


    if (DEBUG) {
	fprintf(This->logfp, "Set Window \nNPP %p\n", instance);
	fprintf(This->logfp, "This %p\n", This);
	fprintf(This->logfp, "This->state %i\n", This->state);
	printf("SW %p\n", instance);
    }


    if (This->window != (Window) window->window) {
	if (DEBUG)
	    fprintf(This->logfp,
		    "Hey we got a new window! old: 0x%x    new 0x%x\n",
		    This->window, window->window);
    }

    if (This->state < STATE_WINDOWSET) {
	if (DEBUG) {
	    fprintf(This->logfp, "Size: %d %d %x\n", window->x, window->y,
		    window->window);
	    fprintf(This->logfp, "Size: %dx%d \n", window->width,
		    window->height, window->window);
	}

	ws = (NPSetWindowCallbackStruct *) window->ws_info;

	netscape_widget =
	    XtWindowToWidget((Display *) ws->display,
			     (Window) window->window);

	XtAddCallback(netscape_widget, XtNdestroyCallback, DestroyCB,
		      (XtPointer) This);
	XtAddEventHandler(netscape_widget, ExposureMask, FALSE,
			  (XtEventHandler) RedrawCB, This);

	This->display = ws->display;
	This->window = (Window) window->window;
	This->netscape_widget = netscape_widget;
	This->width = window->width;
	This->height = window->height;
	This->state = STATE_WINDOWSET;


    } else {
	if (This->setwindow == 0) {
	    if (DEBUG)
		fprintf(This->logfp, "back in SetWindow\n");
	    if (DEBUG)
		fprintf(This->logfp, "New Size: %ix%i\n", window->width,
			window->height);
	    if (This->mode == NP_EMBED) {
		This->embed_width = window->width;
		This->embed_height = window->height;
	    }
	    This->setwindow++;

	    if (This->state != STATE_GETTING_PLAYLIST) {
		This->state = STATE_GETTING_PLAYLIST;
		if (This->href != NULL) {
		    if (strncmp(This->href, "mms:", 4) == 0) {
			sprintf(This->td->list->url, "%s", This->href);
			This->state = STATE_STARTED_PLAYER;
			This->td->list->mmsstream = 1;
			Redraw(This->netscape_widget, instance, NULL);
		    }
		} else {
		    if (This->url != NULL) {
			if (strncmp(This->url, "mms:", 4) == 0) {
			    sprintf(This->td->list->url, "%s", This->url);
			    This->state = STATE_STARTED_PLAYER;
			    This->td->list->mmsstream = 1;
			    Redraw(This->netscape_widget, instance, NULL);
			}
		    }
		}
	    }
	}
    }

    if (DEBUG) {
	fprintf(This->logfp, "Exiting SetWindow\n");
	fflush(This->logfp);
    }

    return NPERR_NO_ERROR;
}

void
NPP_URLNotify(NPP instance, const char *url, NPReason reason,
	      void *notifyData)
{
    PluginInstance *This;
    int playable;
    Node *n;

    if (instance == NULL)
	return;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	printf("NPN_URLNotify called\n");


    if (reason == NPRES_DONE) {
	playable = 0;
	pthread_mutex_lock(&(This->playlist_mutex));

	n = This->td->list;
	while (n != NULL) {
	    if (n->play == 1)
		playable = 1;
	    n = n->next;
	}

	if (playable) {
	    if (This->threadsignaled == 0) {
		pthread_cond_signal(&(This->playlist_complete_cond));
		This->threadsignaled = 1;
	    }
	}

	pthread_mutex_unlock(&(This->playlist_mutex));
    }
}

NPError
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream * stream, NPBool seekable, uint16 * stype)
{
    NPByteRange range;
    PluginInstance *This;
    Node *n;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG) {
	fprintf(This->logfp, "NewStream Requested\n");
	printf("New Stream requested state: %i\n", This->state);
	printf("URL: %s\n", stream->url);
    }

    if (This->mode == NP_FULL) {
	if (DEBUG)
	    fprintf(This->logfp, "NP_ASFILEONLY\n");
	*stype = NP_ASFILEONLY;
    } else {
	//pthread_mutex_lock(&playlist_mutex);
	if (This->state <= STATE_GETTING_PLAYLIST) {
	    This->state = STATE_GETTING_PLAYLIST;
	    Redraw(This->netscape_widget, instance, NULL);
	}
	if (DEBUG)
	    fprintf(This->logfp, "NP_NORMAL\n");
	*stype = NP_NORMAL;
	//pthread_mutex_unlock(&playlist_mutex);
    }

    return NPERR_NO_ERROR;

}


/* PLUGIN DEVELOPERS:
 *	These next 2 functions are directly relevant in a plug-in which
 *	handles the data in a streaming manner. If you want zero bytes
 *	because no buffer space is YET available, return 0. As long as
 *	the stream has not been written to the plugin, Navigator will
 *	continue trying to send bytes.  If the plugin doesn't want them,
 *	just return some large number from NPP_WriteReady(), and
 *	ignore them in NPP_Write().  For a NP_ASFILE stream, they are
 *	still called but can safely be ignored using this strategy.
 */

int32 STREAMBUFSIZE = 0X0FFFFFFF;	/* If we are reading from a file in NPAsFile
					 * mode so we can take any size stream in our
					 * write call (since we ignore it) */
int32 NPP_WriteReady(NPP instance, NPStream * stream)
{
    PluginInstance *This;
    Node *n;

    if (instance == NULL)
	return -1;

    if (DESTROYED) return -1;
    This = (PluginInstance *) instance->pdata;
    if (This->state == STATE_PLAY_CANCELLED) return -1;

    if (DEBUG) {
	fprintf(This->logfp, "in WriteReady. stream->url: %s\n",
		stream->url);
	fprintf(This->logfp, "in WriteReady, state = %i\n", This->state);
	fflush(This->logfp);

	fprintf(This->logfp, "Setting of This->forced = %i\n",
		This->forced);
    }

    if (This->mode == NP_FULL) {
	return STREAMBUFSIZE;
    } else {

	if (This == NULL)
	    return -1;
	if (This->td == NULL)
	    return -1;

	pthread_mutex_lock(&(This->playlist_mutex));
	if (strlen(stream->url) >= 1023)
	    return -1;
	n = This->td->list;
	while (n != NULL) {
	    if (strlen(n->url) != 0) {
		if (strcmp(n->url, stream->url) == 0) {
		    break;
		} else {
		    if (strstr(stream->url, n->url) != NULL) {
			break;
		    }
		}
	    } else {
		sprintf(n->url, "%s", stream->url);
		break;
	    }
	    n = n->next;
	}

	if (n != NULL) {
	    if (strlen(n->fname) == 0) {
		sprintf(n->fname, "%s",
			tempnam("/tmp", "mplayerplug-inXXXXXX"));
		if (DEBUG)
		    printf("WR tempname: %s\n", n->fname);

	    }
	    pthread_mutex_unlock(&(This->playlist_mutex));
	    return STREAMBUFSIZE;
	} else {
	    if (DEBUG)
		printf("didn't find the node in the playlist\n");
	    n = newNode();
	    sprintf(n->url, "%s", stream->url);
	    sprintf(n->fname, "%s",
		    tempnam("/tmp", "mplayerplug-inXXXXXX"));
	    addToEnd(This->td->list, n);
	    pthread_mutex_unlock(&(This->playlist_mutex));
	    return STREAMBUFSIZE;
	}
    }
}


int32
NPP_Write(NPP instance, NPStream * stream, int32 offset, int32 len,
	  void *buffer)
{
    PluginInstance *This;
    int ret;
    FILE *fp;
    Node *n, *hrefnode;

    if (instance == NULL)
	return -1;

    if (DESTROYED) return -1;
    This = (PluginInstance *) instance->pdata;
    if (This->state == STATE_PLAY_CANCELLED) return -1;

    if (DEBUG)
	fprintf(This->logfp, "in Write. stream->url: %s\n", stream->url);
    if (DEBUG)
	fprintf(This->logfp, "in Write, state = %i\n", This->state);
    if (DEBUG)
	fflush(This->logfp);

    if (This == NULL)
	return -1;
    if (This->td == NULL)
	return -1;

    pthread_mutex_lock(&(This->playlist_mutex));
    // find the node that matches the URL and open the file name in it.
    if (strlen(stream->url) >= 1023)
	return -1;
    n = This->td->list;
    while (n != NULL) {
	if (strcmp(n->url, stream->url) == 0) {
	    break;
	} else {
	    if (strstr(stream->url, n->url) != NULL) {
		break;
	    }
	}
	n = n->next;
    }

    if (n == NULL) {
	pthread_mutex_unlock(&(This->playlist_mutex));
	return -1;
    } else {
	if (n->status != STATE_CANCELLED) {
	    fp = fopen(n->fname, "a+");
	    fseek(fp, offset, SEEK_SET);
	    ret = fwrite(buffer, 1, len, fp);
	    n->bytes = n->bytes + ret;
	    n->retrieved = 1;
	    fclose(fp);

	    if (This->href != NULL) {
		if (This->hrefrequested == 0) {
		    if (DEBUG)
			printf("requesting href\n");
		    This->hrefrequested = 1;
		    hrefnode = newNode();
		    strcpy(hrefnode->url, This->href);
		    addToEnd(This->td->list, hrefnode);
		    NPN_GetURL(instance, This->href, NULL);
		}
	    }


	    if (n->status != STATE_DOWNLOADED_ENOUGH) {
		if (This->baseurl == NULL)
		    This->baseurl = getURLBase(n->url);

		if (DEBUG)
		    printf("calling buildPlaylist with filename %s\n",
			   n->fname);
		if (!isMms(n->url)) {
		    buildPlaylist(instance, n->fname, n);
		} else {
		    n->mmsstream = 1;
		}
	    }
	    if (n->play == 1) {
		if (n->mmsstream == 1) {
		    if (This->threadsignaled == 0) {
			if (DEBUG)
			    printf("signalling player from write\n");
			pthread_cond_signal(&(This->playlist_complete_cond));
			This->threadsignaled = 1;
		    }
		} else {
		    if (n->bytes > (This->cachesize * 1024)) {
			if (This->threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player from write\n");
			    pthread_cond_signal(&(This->playlist_complete_cond));
			    This->threadsignaled = 1;
			}

		    }
		}
	    }

	    n->status = STATE_DOWNLOADED_ENOUGH;
	    pthread_mutex_unlock(&(This->playlist_mutex));
	    return ret;
	}
    }
}


NPError NPP_DestroyStream(NPP instance, NPStream * stream, NPError reason)
{
    PluginInstance *This;
    int playable;
    Node *n;
    int ret;

    if (instance == NULL)
	return NPERR_INVALID_INSTANCE_ERROR;
    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	fprintf(This->logfp, "in destroy stream\n");
    if (DEBUG)
	fflush(This->logfp);

    if (DEBUG)
	printf("NPP_DestroyStream called %i\n URL: %s\n", reason,
	       stream->url);


    if (reason == NPRES_DONE) {
	playable = 0;
	pthread_mutex_lock(&(This->playlist_mutex));
	if (strlen(stream->url) >= 1023)
	    return;
	n = This->td->list;
	while (n != NULL) {
	    if (strcmp(n->url, stream->url) == 0) {
		break;
	    } else {
		if (strstr(stream->url, n->url) != NULL) {
		    break;
		}
	    }
	    n = n->next;
	}
	if (n != NULL) {
	    if (n->play == 1)
		playable = 1;
	    printf("playable %i \n", playable);
	    if (playable) {
		if (This->threadsignaled == 0) {
		    if (DEBUG)
			printf("signalling player\n");
		    pthread_cond_signal(&(This->playlist_complete_cond));
		    This->threadsignaled = 1;
		}
	    } else {
		// look for mmsstream
		n = This->td->list;
		while (n != NULL) {
		    if (n->mmsstream) {
			if (This->threadsignaled == 0) {
			    if (DEBUG)
				printf("signalling player\n");
			    pthread_cond_signal(&(This->playlist_complete_cond));
			    This->threadsignaled = 1;
			}
			break;
		    }
		    n = n->next;
		}
	    }
	}
	pthread_mutex_unlock(&(This->playlist_mutex));
    }

    return NPERR_NO_ERROR;
}


void NPP_StreamAsFile(NPP instance, NPStream * stream, const char *fname)
{
    PluginInstance *This;
    FILE *infile;
    char c;
    char s[MAX_BUF_LEN];
    int i = 0;
    Node *n;

    if (instance == NULL)
	return;

    This = (PluginInstance *) instance->pdata;

    if (DEBUG)
	printf("SaF %p\n", instance);


    if (DEBUG) {
	fprintf(This->logfp, "in stream as file. stream->url: %s\n",
		stream->url);
	fprintf(This->logfp, "in stream as file. fname: %s\n", fname);
	printf("in stream as file. stream->url: %s\n", stream->url);
	printf("in stream as file. fname: %s\n", fname);
    }
    This->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
    This->td->list = newNode();
    This->td->instance = instance;
    This->url = strdup(stream->url);
    This->fname = strdup(fname);
    Redraw(This->netscape_widget, instance, NULL);
    return;
}


void NPP_Print(NPP instance, NPPrint * printInfo)
{
    return;
}

static void DestroyCB(Widget widget, XtPointer client_data,
		      XtPointer call_data)
{
    PluginInstance *This = (PluginInstance *) client_data;
    char cmd[1000];

    if (DEBUG)
	printf("window destroy callback\n");
    if (widget == This->netscape_widget) {
        if (This == NULL) return;
	DESTROYED = 1;
        pthread_mutex_lock(&(This->playlist_mutex));
	XtRemoveCallback(This->netscape_widget, XtNdestroyCallback,
			 DestroyCB, (XtPointer) This);
	XtRemoveEventHandler(This->netscape_widget, ExposureMask, FALSE,
			     (XtEventHandler) RedrawCB, This);
	This->netscape_widget = NULL;
	This->state = STATE_PLAY_CANCELLED;

	sprintf(cmd,
		"ps -eaf | grep mplayer | awk '{print $2}' | xargs kill");
	system(cmd);
	//usleep(500);
	deleteList(This->td->list);
	This->td->list == NULL;

	if (This->td != NULL) {
	    NPN_MemFree(This->td);
	    This->td = NULL;
	}
	if (This->baseurl != NULL) {
	    NPN_MemFree(This->baseurl);
	    This->baseurl = NULL;
	}
	if (This->hostname != NULL) {
	    NPN_MemFree(This->hostname);
	    This->hostname = NULL;
	}


	This->td = (ThreadData *) NPN_MemAlloc(sizeof(ThreadData));
	This->td->list = newNode();
	This->td->instance = NULL;
        pthread_mutex_unlock(&(This->playlist_mutex));

    } else {
	if (DEBUG)
	    printf("Widgets don't match %i != %i\n", widget,
		   This->netscape_widget);
    }

}

static void RedrawCB(Widget widget, XtPointer client_data,
		     XtPointer call_data)
{
    char message[1024];

    PluginInstance *This = (PluginInstance *) client_data;
    if (DEBUG)
	printf("redrawing window\n");

    if (DESTROYED) return;

    if (This != NULL) {
	if (This->lastmessage != NULL) {
	    if (strlen(This->lastmessage) != 0) {
		if (This->td->instance != NULL) {
		    if (This->state >= STATE_GETTING_PLAYLIST) {
			strcpy(message, This->lastmessage);
			DrawUI(This->netscape_widget, This->td->instance,
			       message);
		    }
		}
	    }
	}
    }
}

void Redraw(Widget w, NPP instance, XEvent * event)
{

    PluginInstance *This;
    Node *node;
    int i = 0;
    char xval[32], yval[32], yval2[32];

    if (DEBUG)
	printf("Redraw\n");
    if (instance == NULL) {
	if (DEBUG)
	    printf("Our instance is null!\n");
	return;
    }
    // structures

    This = (PluginInstance *) instance->pdata;
    if (This->redrawcalled == 1)
	return;

//    pthread_mutex_lock(&playlist_mutex);

    This->redrawcalled = 1;
    This->td->instance = instance;
    This->td->w = w;

    if (This->download == 1) {
	DrawUI(w, instance, "Downloading Media...");
    } else {
	DrawUI(w, instance, "Loading Media...");
    }


    if (This->mode == NP_FULL) {
	sprintf(This->td->list->url, "%s", This->url);
	sprintf(This->td->list->fname, "%s", This->fname);
	sprintf(xval, "%i", This->width);
	sprintf(yval, "%i", This->height);
	sprintf(yval2, "%i", This->height - 16);
	This->td->list->delete = 0;	// if we delete this we mess up the Mozilla cache
	This->maintain_aspect = 0;
	if (This->fileonly_embed == 0)
	    This->noembed = 1;
    } else {
	if (This->href) {
	    sprintf(This->td->list->url, "%s", This->href);
        } else if (This->fname) {
            sprintf(This->td->list->url, "%s", This->fname);
	} else {
	    sprintf(This->td->list->url, "%s", This->url);
	}
	sprintf(xval, "%i", This->embed_width);
	sprintf(yval, "%i", This->embed_height);
	sprintf(yval2, "%i", This->embed_height - 16);
    }

    This->baseurl = getURLBase(This->td->list->url);
    This->hostname = getURLHostname(This->td->list->url);

    sprintf(This->td->argv[i++], "%s", This->player);

    if (This->novop == 1) {
	sprintf(This->td->argv[i++], "-vop");
	sprintf(This->td->argv[i++], "null");
    } else {
	if (This->vopopt != NULL) {
	    sprintf(This->td->argv[i++], "-vop");
	    sprintf(This->td->argv[i++], "%s", This->vopopt);
	    if (strncmp(This->vopopt, "scale=", 6) == 0) {
		sprintf(This->td->argv[i++], "-fs");
	    }
	}
    }

    if (This->noembed == 0) {
	if (This->window != 0) {
	    sprintf(This->td->argv[i++], "-wid");
	    sprintf(This->td->argv[i++], "0x%x", This->window);
	}
    }

    if (DEBUG)
	printf("window defined %s \n", This->td->argv[i - 1]);

    if ((This->embed_width == 0)
	&& (This->mode == NP_EMBED)) {

    } else {
	if (This->scaling == 1) {
	    if (This->use_gmplayer > 0) {
		sprintf(This->td->argv[i++], "-screenw");
		sprintf(This->td->argv[i++], "%s", xval);
		sprintf(This->td->argv[i++], "-screenh");
		sprintf(This->td->argv[i++], "%s", yval);
		if (This->maintain_aspect == 1) {
		    sprintf(This->td->argv[i++], "-xy");
		    sprintf(This->td->argv[i++], "%s", xval);
		} else {
		    sprintf(This->td->argv[i++], "-x");
		    sprintf(This->td->argv[i++], "%s", xval);
		}
		if (This->use_gmplayer == 2) {
		    if (This->maintain_aspect == 0) {
			sprintf(This->td->argv[i++], "-y");
			sprintf(This->td->argv[i++], "%s", yval2);
		    }
		    sprintf(This->td->argv[i++], "-skin");
		    sprintf(This->td->argv[i++], "mini");
		    if (This->noembed == 0) {
			sprintf(This->td->argv[i++], "-guiwid");
			sprintf(This->td->argv[i++], "0x%x", This->window);
		    }
		} else {
		    if (This->maintain_aspect == 0) {
			sprintf(This->td->argv[i++], "-y");
			sprintf(This->td->argv[i++], "%s", yval);
		    }
		}
	    } else {
		if (This->maintain_aspect == 1) {
		    sprintf(This->td->argv[i++], "-xy");
		    sprintf(This->td->argv[i++], "%s", xval);
		} else {
		    sprintf(This->td->argv[i++], "-x");
		    sprintf(This->td->argv[i++], "%s", xval);
		    sprintf(This->td->argv[i++], "-y");
		    sprintf(This->td->argv[i++], "%s", yval);
		}
	    }
	}
    }

    if (This->rtsp_use_tcp == 1) {
	sprintf(This->td->argv[i++], "-rtsp-stream-over-tcp");
    }

    if (This->cachesize > 0) {
	sprintf(This->td->argv[i++], "-cache");
	sprintf(This->td->argv[i++], "%i", This->cachesize);
    } else {
	sprintf(This->td->argv[i++], "-nocache");
    }

    if (This->loop == 1) {
	sprintf(This->td->argv[i++], "-loop");
	sprintf(This->td->argv[i++], "0");
    }

    if (This->vo) {
	sprintf(This->td->argv[i++], "-vo");
	sprintf(This->td->argv[i], "%s", This->vo);
	if (strcmp(This->td->argv[i++], "x11") == 0) {
	    sprintf(This->td->argv[i++], "-zoom");	/* -vo x11 needs -zoom for rescaling */
	}
    }

    if (This->ao) {
	sprintf(This->td->argv[i++], "-ao");
	sprintf(This->td->argv[i++], "%s", This->ao);
    }

    if (This->output_display) {
	sprintf(This->td->argv[i++], "-display");
	sprintf(This->td->argv[i++], "%s", This->output_display);
    }

    sprintf(This->td->argv[i++], "-osdlevel");
    sprintf(This->td->argv[i++], "%i", This->osdlevel);

    /* some extra flags for mplayer */
    sprintf(This->td->argv[i++], "-nojoystick");	/* annoying... */
    sprintf(This->td->argv[i++], "-nofs");	/* no full screen */

    // if we don't do this next part, using use-gui=[yes|mini] has annoying popups
    if (strcasecmp(This->player,"gmplayer") == 0 ) {
	sprintf(This->td->argv[i++], "-really-quiet");
    } else {
	sprintf(This->td->argv[i++], "-quiet");
    }
    sprintf(This->td->argv[i++], "");

    // ok we have the command line setup
    if (DEBUG)
	printf("ready to setup threads\n");

    pthread_attr_init(&(This->thread_attr));
    pthread_attr_setdetachstate(&(This->thread_attr), PTHREAD_CREATE_DETACHED);
    pthread_cond_init(&(This->playlist_complete_cond), NULL);
    This->current_node = NULL;

    if (This->mode == NP_FULL) {

	if (DEBUG)
	    printf("creating thread - NP_FULL\n");
	if (This->state < STATE_PLAYLIST_COMPLETE)
	    This->state = STATE_PLAYLIST_COMPLETE;
//      pthread_mutex_unlock(&playlist_mutex);
	pthread_create(&(This->player_thread), &(This->thread_attr), playPlaylist,
		       (void *) This->td);

    } else {

	if (DEBUG)
	    printf("creating thread - NP_EMBED\n");
	DrawUI(w, instance, "Getting playlist...");
	if (This->state < STATE_GETTING_PLAYLIST)
	    This->state = STATE_GETTING_PLAYLIST;
//      pthread_mutex_unlock(&playlist_mutex);
	if (DEBUG)
	    printf("creating player thread\n");
	pthread_create(&(This->player_thread), &(This->thread_attr), playPlaylist,
		       (void *) This->td);
	// recommended slight pause
	usleep(1);
    }
    if (DEBUG)
	printf("MAIN THREAD DONE\n");
}
